%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ALGORITMUS MAX. TOKU V SITI ( GOLDBERG )
% Ondrej Dusek (33, Patek 9:00 S1)
%
% Zjisti max. tok v siti za pouziti Goldbergova
% algoritmu. Sit je reprezentovana v databazi, 
% ve formatu "hrana( Z, Do, Kapacita )."
% ( zadano pro kazdou hranu ), navic musi byt 
% definovan "zdroj( Z )." a "stok( S )."
%
% po zavolani "max_flow." program vypise na std.
% vystup max. tok kazdou hranou ve formatu:
% "Z -> Do : Tok"


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ziskani seznamu vrcholu pro inicializaci
%

% test prislusnosti prvku v seznamu 
% clen( +Prvek, +Seznam )
clen( X, [ X | _ ] ).
clen( X, [ Z | XS ] ):- Z \= X, clen( X, XS ).

% slepi 2 (nesetridene) seznamy
% slep( +Seznam1, +Seznam2, -Slepeny )
slep( [], Y, Y ).
slep( [ X | XS ], Y ,  S ):- clen( X, Y ), !, slep( XS, Y, S ).
slep( [ X | XS ], Y , [ X | S ] ):- slep( XS, Y, S ).

% vyda seznam vrcholu 
% vrcholy( -Vrcholy ).
vrcholy( B ):-setof( X, C^A^hrana( A, X, C ), B1 ), setof( Y, C^A^hrana( Y, A, C ), B2 ), slep( B1, B2, B ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% inicializace
%

% prida k vrcholum prebytek = 0, vrcholy primo spojene se zdrojem = prebytek - kap. hrany ze zdroje
% init_v( +Vrcholy, -VrcholySVyskou ).
init_v( [], [] ).
init_v( [ X | XS ], [ X-D | IXS ] ):-zdroj( V ), hrana( V, X, D), !, init_v( XS, IXS ).
init_v( [ X | XS ], [ X-0 | IXS ] ):-init_v( XS, IXS ).


% spocita pocet prvku v seznamu
% spocti_pocet( +Seznam, -Pocet ).
spocti_pocet( [], 0 ).
spocti_pocet( [ _ | XS ], A ):-spocti_pocet( XS, A1 ), A is A1 + 1.

% inicializuje tok
% init_preflow( +PocetVrcholu, +NulovyTok, -PocPreFlow ).
init_preflow( _, [], [] ).
init_preflow( N, [ V-W-0-0 | XS ], [ V-W-0-N | IXS ] ):-zdroj( V ), !, hrana( V, W, _ ), 
    init_preflow( N, XS, IXS ).
init_preflow( N,  [ V-W-0-0 | XS ], [ V-W-C-0 | IXS ] ):-hrana( V, W, C ), init_preflow( N, XS, IXS ).

% vyrobi seznam hran s pocatecnim tokem & poc. rozdilem vysky
% init_e( -InitHrany ).
init_e( N, E ):- setof( X-Y-0-0, C^hrana( X, Y, C ), E1 ),  init_preflow( N, E1, E ).

% inicializace site
% init( +Vrcholy, -InitVrcholy, -InitHrany )
init( B, V, E ):- init_v( B, V ), spocti_pocet( V, N ), init_e( N, E ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% nalezeni vrcholu s akt. kladnym prebytkem
%

% nalezne vrchol s aktualne kladnym prebytkem
% najdi_vdelta( +Vrcholy, -Vrchol, -Zbytek )
najdi_vdelta( [ X-_ | V ], X0-P0 ):- ( stok( X ) ; zdroj( X ) ), !, najdi_vdelta( V, X0-P0 ).
najdi_vdelta( [ X-P | _ ], X-P ):- P > 0.
najdi_vdelta( [ _-P | V ], X0-P0 ):- P = 0, najdi_vdelta( V, X0-P0 ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% nalezeni hrany pro prevedeni prebytku
%

% najde hranu vedouci z V do W, t.z r(h) > 0, vyska(V) > vyska( W ).
% najdi_hranu( +Vrchol, +Ehrany, -Hrana, -Zbytekhran, -Zpatky )
najdi_hranu( X, [ V-W-R0-DH0 | XS ], X1-Y1-R-DH, [ V-W-R0-DH0 | Z ], B ):- V \= X, W \= X,
    najdi_hranu( X, XS, X1-Y1-R-DH, Z , B), !.
najdi_hranu( X, [ V-W-R0-DH0 | XS ], X1-Y1-R-DH, [ V-W-R0-DH0 | Z ], B ):- hrana( V, W, C ),
    (( X \= V, R0 = C); (X \= W, R0 = 0 )), najdi_hranu( X, XS, X1-Y1-R-DH, Z, B ), !.
najdi_hranu( X, [ V-W-R0-DH0 | XS ], X1-Y1-R-DH, [ V-W-R0-DH0 | Z ], B ):- ((X = V, DH0 =< 0) ; ( X = W, DH0 >= 0 )),
    najdi_hranu( X, XS, X1-Y1-R-DH, Z, B ), !.
najdi_hranu( X, [ X-Y-R-DH | XS ], X-Y-R-DH, XS, front ):- R > 0, DH > 0, !.
najdi_hranu( X, [ Y-X-R-DH | XS ], Y-X-R-DH, XS, back ):- hrana( Y, X, C ), R < C, DH < 0, !.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% prevedeni prebytku
%


% najde minimum 2 cisel
% minimum( +Cislo1, +Cislo2, -Minimum )
minimum( X, Y, X ):- X =< Y.
minimum( X, Y, Y ):- X > Y.


% vybere ze seznamu vrcholu V vrchol X s prebytkem, zbytek vrcholu ulozi do Zbytek
% najdi_vrchol( +Vrchol, +Vrcholy, -Vrchol&prebytek, -Zbytek ).
najdi_vrchol( X, [ Y-P0 | V ], X-P, [ Y-P0 | Z ] ):- X \= Y, najdi_vrchol( X, V, X-P, Z ).
najdi_vrchol( X, [ X-P | V ], X-P, V ).


% prevede prebytek z poc. do koncoveho vrcholu hrany 
% preved_delta( +Hrana, +Vrcholy, +Zpatky, -Hrana, -Vrcholy )
preved_delta( X-Y-R-DH, V, front, X-Y-R1-DH, [ X-PX1, Y-PY1 | V2 ] ):- najdi_vrchol( X, V, X-PX, V1 ), 
    najdi_vrchol( Y, V1, Y-PY, V2 ), minimum( R, PX, M ), PX1 is PX - M, PY1 is PY + M, R1 is R - M.
preved_delta( X-Y-R-DH, V, back, X-Y-R1-DH, [ X-PX1, Y-PY1 | V2 ] ):- hrana( X, Y, C ), 
    najdi_vrchol( X, V, X-PX, V1 ), najdi_vrchol( Y, V1, Y-PY, V2 ), RB is C - R, minimum( PY, RB, M ), 
    PX1 is PX + M, PY1 is PY - M, R1 is R + M.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% zvednuti vrcholu
%

% "zvedne vrchol" - vsem hranam vedoucim do/ z nej upravi rozdil vysky
% ( misto "vysky vrcholu" (Goldberg)  se pro kazdou hranu pouziva "rozdil vysky" )
% zvedni_v( +Vrchol, +Hrany, -UprHrany )
zvedni_v( _, [], [] ).
zvedni_v( V, [ X-Y-R-DH | XS ], [ X-Y-R-DH | ZXS ] ):- V \= X, V \= Y, zvedni_v( V, XS, ZXS ).
zvedni_v( V, [ V-Y-R-DH | XS ], [ V-Y-R-DH1 | ZXS ] ):- DH1 is DH + 1, zvedni_v( V, XS, ZXS ).
zvedni_v( V, [ X-V-R-DH | XS ], [ X-V-R-DH1 | ZXS ] ):- DH1 is DH - 1, zvedni_v( V, XS, ZXS ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Goldberg
%

% Goldberguv algoritmus 
% goldberg( +Vrcholy, +Hrany, -Tok )
goldberg( V, E, FL ):- najdi_vdelta( V, X-_ ), !, 
	( najdi_hranu( X, E, H, EZ, ZP ) ->    
	    ( preved_delta( H, V, ZP, H1, V1 ), E2 = [ H1 | EZ ] )
	    ; ( zvedni_v( X, E, E2 ), V1 = V )
	),
	goldberg( V1, E2, FL ).
% alg. skonci pokud predikat najdi_vdelta neuspeje
goldberg( _, E, E ).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% MAIN
%

% prevede sit rezerv na vystupni format - tok a vypise
% write_flow( +Sit, -Tok ).
write_flow( [] ).
write_flow( [ X-Y-R-_ | XS ] ):- hrana( X, Y, C ), F is C - R,  write( X -> Y ),
	write( ' :   ' ), write(  F ), write( '\n') , write_flow( XS ).

% hlavni predikat
max_flow:-vrcholy( B ),init( B , V, E ), goldberg( V, E, E1 ), write( '\n' ), write_flow( E1).
